
#-----------------------------------
# Imports
#-----------------------------------
import numpy as np
import importlib
import yaml
from typing import Dict

from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.core.plci_core_data_type import DataType, get_data_type
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData, CompositeBoxData, CodeBoxData, MultiPhysicsSimulationBoxData
from muphyn.packages.core.box_library.abstract_box_library_importer import AbstractBoxLibraryImporter

#-----------------------------------
# Class
#-----------------------------------

class BoxLibraryImporter_1_1_0 (AbstractBoxLibraryImporter) :
    """
    Est la classe V 1.1.0 d'importeur de boxes.

    What's new ?
    ------------
     - Importer is now able to link parameter value changed event to external functions
    """
    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        super().__init__()
        
    # -------------
    # Methods
    # -------------

    def import_box (self, path : str, file_name : str, absolute_yaml_file : str, boxes : Dict[str, AbstractBoxData]) -> Dict[str, AbstractBoxData] :
        """Est la méthode pour charger des boxes depuis des fichiers."""
        
        # Open File
        with open(absolute_yaml_file) as yaml_file_data :

            # Open YAML File
            file_data = yaml.load(yaml_file_data, Loader = yaml.FullLoader)

            if 'box' in file_data :
                # Extract Box data
                box_data = file_data['box']

                # Filter Box based library name
                if box_data['library'].startswith('Boxes') :
                    
                    # Build full box name
                    library_name = box_data['library'] + "." + box_data['name']

                    # If box already exists
                    if library_name in boxes :
                        # If existing box has a more recent version → Abort import current box
                        if boxes[library_name].version >= box_data['version']:
                            return None
                    
                    if box_data['type'] in ['code', 'multiphysics-simulation'] :
                        return {'library_name' : library_name, 'box_data' : self._import_code_box(library_name, box_data, path + '/' + file_name + '.py')}

                    if box_data['type'] == 'multiphysics-simulation' :
                        return {'library_name' : library_name, 'box_data' : self._import_multiphysics_simulation_box(library_name, box_data, path + '/' + file_name + '.py')}

                    elif box_data['type'] == 'composite' :
                        return {'library_name' : library_name, 'box_data' : self._import_composite_box(library_name, box_data)}
                    
                elif box_data['library'].startswith('Electrical'):
                    
                    library_name = box_data['library'] + "." + box_data['name']


    def _import_composite_box (self, library_name, box_data) -> CompositeBoxData :
        """Permet de générer un instanciateur de box composite."""

        if box_data['inputs'] == 'None':
            inputs = []
        else :
            inputs = box_data['inputs']

        if box_data['outputs'] == 'None':
            outputs = []
        else :
            outputs = box_data['outputs']

        return CompositeBoxData (
            box_data['name'],
            box_data['library'],
            box_data['creator'],
            box_data['date_creation'],
            box_data['version'],
            box_data['signals'],
            box_data['boxes'],
            inputs,
            outputs,
        )   

    def _import_code_box (self, library_name, box_data, absolute_py_file) -> CodeBoxData :
        """Permet de générer un instanciateur de box de code."""

        # Importations des méthodes du fichier python.
        spec = importlib.util.spec_from_file_location(library_name, absolute_py_file)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)

        # Si la méthode init n'éxiste, pas création de celle-ci.
        if "box_init_method" in box_data and box_data['box_init_method'] != "None" : 
            init_method = getattr(foo, box_data['box_init_method'])
        else :
            init_method = lambda box, scheduler : ...

        # Si la fonction n'éxiste, pas création de celle-ci
        if "box_function" in box_data and box_data['box_function'] != "None" : 
            box_function = getattr(foo, box_data['box_function'])
        else :
            box_function = lambda box, event : ...

        # Si la méthode end n'éxiste, pas création de celle-ci
        if "box_end_method" in box_data and box_data['box_end_method'] != "None" : 
            end_method = getattr(foo, box_data['box_end_method'])
        else :
            end_method = lambda box : ...

        inputs_groups = {}
        if "inputs" in box_data and box_data['inputs'] != 'None' :
            inputs = box_data['inputs']
            for input_index, input_group in enumerate(inputs):
                # Get input name
                input_group_name = f"Unnamed:{input_index}" if "name" not in input_group else input_group["name"]

                # Build input group
                if input_group_name not in inputs_groups:
                    # handling input values
                    inputs_groups[input_group_name] = BoxLibraryImporter_1_1_0.build_io_group(input_group)
                else:
                    LogManager().error(f"Duplicate input name detected for box {library_name} : {input_group_name}")
        else :
            inputs = {}

        outputs_groups = {}
        if "outputs" in box_data and box_data['outputs'] != 'None' :
            outputs = box_data['outputs']
            for output_index, output_group in enumerate(outputs):
                # Get output name
                output_group_name = f"Unnamed:{output_index}" if "name" not in output_group else output_group["name"]

                # Build output group
                if output_group_name not in outputs_groups:
                    # handling output values
                    outputs_groups[output_group_name] = BoxLibraryImporter_1_1_0.build_io_group(output_group)
                else:
                    LogManager().error(f"Duplicate output name detected for box {library_name} : {output_group_name}")
        else :
            outputs = {}

        if "params" in box_data and box_data['params'] != 'None' :
            params = box_data['params']

            # Handle params
            for param_name, param_description in params.items():

                if "type" in param_description:
                    # Handle Choice type
                    if param_description["type"] == "choice":
                        # Handle default index set
                        if "value" in param_description and "choices" in param_description:
                            index = param_description["value"]
                            choices = param_description["choices"]
                            if type(index) == int and index >= -1 and index < len(param_description["choices"]):
                                if type(choices) == list:
                                    param_description["value"] = choices[index]
                                elif type(choices) == dict:
                                    key = list(choices.keys())[index]
                                    param_description["value"] = choices[key]
                                else:
                                    param_description["value"] = None
                            else:
                                param_description["value"] = None

                        else:
                            param_description["value"] = None
                else:
                    LogManager().error(f"{library_name}: No Data type set for {param_name}")

        else :
            params = {}

        # Callbacks
        for param, attributes in params.items():
            if "on_value_changed" in attributes:
                on_value_changed_function_name = attributes["on_value_changed"]
                if hasattr(foo, on_value_changed_function_name):
                    attributes["callback"] = getattr(foo, on_value_changed_function_name)
                else:
                    LogManager().error(f"Function name doesn't exist for param {param}.on_value_changed: {on_value_changed_function_name}")


        if "icon" not in box_data :
            icon = None
        else:
            icon = box_data['icon']

        # Retour de la nouvelle box de code.
        return CodeBoxData(
            box_name = box_data['name'],
            box_library = box_data['library'],
            box_type = box_data["type"],
            wait_for_event = box_data['wait_for_events'],
            wait_for_all_signal_events = box_data['wait_for_all_signal_events'],
            params = params,
            box_init_method_ = init_method, 
            box_function_ = box_function,
            box_end_method_ = end_method, 
            creator = box_data['creator'],
            date_created = box_data['date_creation'],
            version = box_data['version'],
            inputs = inputs_groups,
            outputs = outputs_groups,
            icon = icon
        )

    def _import_multiphysics_simulation_box(self, library_name, box_data, absolute_py_file) -> MultiPhysicsSimulationBoxData:
        code_box_data: CodeBoxData = self._import_code_box(library_name, box_data, absolute_py_file)

        return MultiPhysicsSimulationBoxData.fromCodeBoxData(code_box_data)

    @staticmethod
    def build_io_group(io_group: dict) -> dict:
        # Type 
        io_group_type = DataType.UNDIFINED if "type" not in io_group else get_data_type(io_group["type"])

        # Is Infinite
        io_group_is_infinite = False if "isInfinite" not in io_group else io_group["isInfinite"]
        if type(io_group_is_infinite) != bool:
            LogManager().error(f"Wrong 'isInfinite' value: '{io_group_is_infinite}' instead of 'True' or 'False'\nThis value has been set to 'False'")
            io_group_is_infinite = False

        # Minimum count
        io_group_minimum_count = 0 if "minimumCount" not in io_group else io_group["minimumCount"]
        if io_group_minimum_count < 0:
            LogManager().error(f"Wrong 'minimumCount' value: '{io_group_minimum_count}' is negative.\nThis value has been set to '0'")
            io_group_minimum_count = 0

        # Maximum count
        int32_info = np.iinfo(np.int32)
        io_group_maximum_count = int32_info.max if "maximumCount" not in io_group else io_group["maximumCount"]
        if io_group_maximum_count < io_group_minimum_count:
            LogManager().error(f"Wrong 'maximumCount' value: '{io_group_maximum_count}' is lower than 'minimumCount' value '{io_group_minimum_count}'.\nThis value has been set to '{io_group_minimum_count+1}'")
            io_group_maximum_count = io_group_minimum_count + 1

        # Count
        io_group_count = None if "count" not in io_group else io_group["count"]
        # If no value has been set
        if io_group_count is None:
            # If io group is 'Infinite' → 2 else if io group is 'Finite' → 0 
            io_group_count = 2 if io_group_is_infinite else 0

        return {
            "type": io_group_type,
            "isInfinite": io_group_is_infinite,
            "minimumCount": io_group_minimum_count,
            "maximumCount": io_group_maximum_count,
            "count": io_group_count
        }


from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibraries
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.interface.dialogs.dialogs_holder import DialogsHolder


boxesLibraries = BoxesLibraries()
schedulersLibraries = SchedulersLibraries()
dialogsHolder = DialogsHolder(None)
dialog = dialogsHolder.show_dialog(name = 'library', modal = False, boxes_libraries = boxesLibraries, solvers_libraries = schedulersLibraries)



from typing import Any
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QHBoxLayout, QComboBox

from muphyn.packages.interface.properties_pages.data_property_widgets.abstract_property_widget import AbstractPropertyWidget
from muphyn.packages.core.utils.log_manager import LogManager

class ChoicePropertyWidget(AbstractPropertyWidget):
    def __init__(self, parameter_to_edit: dict, parent=None, flags=Qt.Widget) -> None:
        super().__init__(parent, flags)

        # Choices
        self._choices: list | dict = parameter_to_edit["choices"] if "choices" in parameter_to_edit else []

        # Init VBoxLayout
        layout = QHBoxLayout()

        # Index
        self._index = -1
        if "value" in parameter_to_edit:
            self._value = parameter_to_edit["value"]
            if type(self._choices) == list and self._value in self._choices:
                self._index = self._choices.index(self._value)
            elif type(self._choices) == dict and self._value in self._choices.values():
                self._index = list(self._choices.values()).index(self._value)

        # Double Spin
        self.combo_box = QComboBox()
        self.combo_box.addItems(self._choices)
        self.combo_box.setCurrentIndex(self._index)
        self.combo_box.currentIndexChanged.connect(self.onNewIndex)

        # Set Layout
        layout.addWidget(self.combo_box, 1)
        self.setLayout(layout)

    def onNewIndex(self, newIndex: int):
        if self._index != newIndex:
            self._index = newIndex
            if type(self._choices) == list:
                self.setValue(self._choices[newIndex])
            elif type(self._choices) == dict:
                key = list(self._choices.keys())[newIndex]

                self.setValue(self._choices[key])

    def setValue(self, new_value: Any):
        if self._value != new_value:
            # Get new value
            self._value = new_value

            # Emit value changed
            self.valueChanged.emit()



from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibraries
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_signal import Signal
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.core.plci_core_scheduler import Scheduler

from muphyn.packages.unit_test.unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries : BoxesLibraries = get_boxes_libraries()
schedulersLibraries : SchedulersLibraries = get_schedulers_library()

print("Creation of the diagram")
# Scheduler parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(0.05, 0.001)

# Signals
scheduler.diagram.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

# Sources
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Step', start_value = 1, stop_value = 0, step_time = 0.035))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Step', start_value = 0, stop_value = 1, step_time = 0.017))

# Reccurence
C = [
        [1/2,   1/4,    1/8,    1/16,   1/32,   1/64],
        [4,     4/3,    4/9,    4/27,   4/81,  4/243],
        [1,     1/5,   1/25,   1/125,  1/625, 1/3125]  
    ]

scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Reccurence', coefficients = C))

# Sink
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = "Reccurence"))
 

scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[0], scheduler.diagram.signals[1])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[2])

# Launch solver
if simulate(scheduler) :
    print("===============================================")
    print("Values : ")
    l = len(scheduler.diagram.boxes[2].params['values'][0])
    for i in range(l) :
        j = l - 1 - i
        print(str(scheduler.diagram.boxes[2].params['values'][0][j]).replace('.', ','))


# General import
from enum import Enum
from typing import Optional

# General imports
import numpy as np

# PyQt imports
from PyQt5.QtCore import QRect,QSize, Qt
from PyQt5.QtGui import QColor, QImage, QPainter, QPixmap, QWheelEvent, QBrush
from PyQt5.QtWidgets import QGraphicsScene, QGraphicsView, QWidget, QApplication

# Project imports
from muphyn.packages.core.utils.log_manager import LogManager

class DrawingScene(QGraphicsScene):

    def __init__(self, parent):
        super().__init__(parent)

class DrawingView(QGraphicsView):

    def __init__(self, scene: QGraphicsScene, parent):

        super().__init__(scene, parent)

        # General View Parameters
        self.setGeometry(self.geometry())
        self.setRenderHint(QPainter.RenderHint.Antialiasing, True)

        # Init zoom
        self.zoomLevel = 5

    def setSize(self, newSize: QSize) -> None:
        # Update geometry
        self.setGeometry(QRect(self.pos(), newSize))

    def wheelEvent(self, event: QWheelEvent):
        if Qt.KeyboardModifier.ControlModifier == QApplication.keyboardModifiers():
            if  event.angleDelta().y() > 0:
                factor = 1.25
                self.zoomLevel += 1
            else:
                if self.zoomLevel > 0:
                    factor = 0.8
                    self.zoomLevel -= 1

            if self.zoomLevel > 0:
                self.scale(factor, factor)

class PatternBuilder:

    class PatternType(Enum):
        CrossPattern = 0
        SquarePattern = 1

    # -------------
    # Constants
    # -------------
    DefaultGridColor = QColor(200, 200, 200)

    @staticmethod
    def buildPixmap(gridSize: int, patternType: PatternType, color: QColor = DefaultGridColor) -> QPixmap:
        if patternType == PatternBuilder.PatternType.CrossPattern:
            return PatternBuilder.buildCrossPattern(gridSize, color)
        elif patternType == PatternBuilder.PatternType.SquarePattern:
            return PatternBuilder.buildSquarePattern(gridSize, color)
        else:
            LogManager().error(f"PatternBuilder.buildPixmap(): Pattern type not supported {patternType}")
            return QPixmap(gridSize, gridSize)

    @staticmethod
    def buildCrossPattern(gridSize: int, color: QColor = DefaultGridColor) -> QPixmap:
        # Convetr color to list
        colorList = [color.red(), color.green(), color.blue()]

        # Init numpy array
        npPattern = np.full((gridSize, gridSize, 3), 255, np.uint8)

        # Change numpy array pattern
        ## Top left corner
        npPattern[0,:3] = colorList
        npPattern[:3,0] = colorList

        ## Top right corner
        npPattern[0,-2:] = colorList

        ## Bottom left corner
        npPattern[-2:,0] = colorList

        bytesPerLine = 3 * gridSize
        return QImage(npPattern.data, gridSize, gridSize, bytesPerLine, QImage.Format_RGB888)

    @staticmethod
    def buildSquarePattern(gridSize: int, color: QColor = DefaultGridColor) -> QPixmap:
        # Convetr color to list
        colorList = [color.red(), color.green(), color.blue()]

        # Init numpy array
        npPattern = np.full((gridSize, gridSize, 3), 255, np.uint8)

        # Change numpy array pattern
        ## Replace first row
        npPattern[0,:] = colorList
        ## Replace first column
        npPattern[1:,0] = colorList

        bytesPerLine = 3 * gridSize
        return QImage(npPattern.data, gridSize, gridSize, bytesPerLine, QImage.Format_RGB888)

class DrawingWidget(QWidget):

    def __init__(self, parent: Optional[QWidget] = None) -> None:

        # Widget flag
        super().__init__(parent)

        # Create board scene
        self.scene = DrawingScene(self)

        # Create Board View
        self.view = DrawingView(self.scene, self)

        # Draw background
        self.view.setBackgroundBrush(QBrush(PatternBuilder.buildPixmap(50, PatternBuilder.PatternType.CrossPattern)))

    def setSize(self, newSize: QSize):
        # Set this item size
        self.setGeometry(QRect(self.pos(), newSize))

        # Update view size
        self.view.setSize(newSize)


#-----------------------------------
# Imports
#-----------------------------------
import copy
import traceback
from typing import Iterable, Any, Dict

from PyQt5.QtCore import QPointF, QSizeF
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData

from muphyn.packages.core.plci_core_data_type import DataType, get_data_type
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel, BoxType
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager

#-----------------------------------
# Class
#-----------------------------------

class BoxModel (AbstractBoxModel) :
    """Est le model graphique des boxes affichées à l'écran."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, library : str, name : str, position : QPointF, size : QSizeF, rotation : float, 
                  can_be_loaded : bool = True, text : str = '', icon: str = None) :
        
        self._library = library
        self._parameters : Dict[str, Any]= {}

        if library == "Boxes.Sources":
            box_type = BoxType.Source
        elif library == "Boxes.Signals":
            box_type = BoxType.Signal
        else:
            box_type = BoxType.Math

        AbstractBoxModel.__init__(self, name, position, size, rotation, text, box_type, icon, None)

        self._can_be_loaded : bool = can_be_loaded
        self.setZValue(1)

    # -------------
    # Properties
    # -------------

    @property
    def library (self) -> str : 
        """Permet de récuperer le nom de la bibliothèque dans laquelle se trouve la box."""
        return self._library

    @library.setter
    def library (self, library_ : str) -> None :
        """Permet de modifier le nom de la bibliothèque dans laquelle se trouve la box."""
        self._library = library_ or ''

    @property
    def can_be_loaded (self) -> bool :
        """Permet de savoir si la box actuelle est bien présente dans la bibliothèques."""
        return self._can_be_loaded

    @can_be_loaded.setter
    def can_be_loaded (self, can_be_loaded_ : bool) -> None : 
        """Permet de modifier le fait que la box actuelle est bien présente dans la bibliothèque."""
        self._can_be_loaded : bool = can_be_loaded_

    @property
    def rendered_text (self) -> str :
        return super().rendered_text

    @rendered_text.setter
    def rendered_text(self, rendered_text_ : str) -> None :

        self._text = rendered_text_

    # -------------
    # Methods
    # -------------

    def set_parameter (self, name : str, value : object) -> None :
        """Permet de modifier la valeur du paramètre.""" 

        old_value = self._parameters[name]['value']
        if value == old_value :
            return

        parameter = self._parameters[name]

        # Set Value
        parameter['value'] = value

        # Handle callback
        if "callback" in parameter:
            callback = parameter["callback"]
            try:
                callback(self)
            except:
                LogManager().error(traceback.format_exc())

        if self._action_param_semaphore:
            return
            
        self.param_changed.emit(self, 'params.'+name, old_value, value)
    
    def get_parameter (self, name : str) -> Dict :
        """Permet de récuperer les données d'un paramètre."""
        return self._parameters[name]

    def get_parameters (self) -> Iterable[str] :
        """Permet de récuperer la liste des paramètres."""
        for key in self._parameters :
            yield key

    def get_parameters_len (self) -> int :
        """Permet de récuperer la taille du tableau de paramètres."""
        return self._parameters.__len__()

    def create_parameter(self, parameter_name : str, parameter_type : DataType, parameter_value : object) -> None :
        """Permet de créer un slot pour le paramètre voulu."""
        self._parameters[parameter_name] = {}
        self._parameters[parameter_name]['type'] = parameter_type
        self._parameters[parameter_name]['value'] = parameter_value

    def to_dict(self) -> dict:
        box_dict = {
            "name": self.name,
            "library": self.library,
            "text": self.text,
            "geometry": {
                "x": self.x(),
                "y": self.y(),
                "width": self.size.width(),
                "height": self.size.height(),
                "rotation": self.rotation()
            },
            "inputs_groups": [],
            "outputs_groups": [],
            "params": {param: self.get_parameter(param)['value'] for param in self.get_parameters()}
        }

        return box_dict

    def reconstructor(self):
        pass

    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def fromBoxData(
            box_data: AbstractBoxData, 
            position: QPointF = QPointF(0, 0), 
            size: QSizeF = QSizeF(AbstractBoxModel.MinimunBoxWidth, AbstractBoxModel.MinimunBoxHeight),
            rotation: float = 0
        ):

        # Handle Geometry value
        ## Size
        if size.width() < AbstractBoxModel.MinimunBoxWidth:
            size.setWidth(AbstractBoxModel.MinimunBoxWidth)
        if size.height() < AbstractBoxModel.MinimunBoxHeight:
            size.setHeight(AbstractBoxModel.MinimunBoxHeight)
        
        ## Rotation
        rotation = ((rotation // 90) % 4) * 90

        # Build Box Model
        box_model = BoxModel(
            library=box_data.box_library,
            name=box_data.box_name,
            position=position,
            size=QSizeF(AbstractBoxModel.MinimunBoxWidth, AbstractBoxModel.MinimunBoxHeight),
            rotation=0, can_be_loaded=True, text="", icon=box_data.icon
        )

        box_model.box_type = box_data.box_type

        # Inputs
        box_model.add_inputs_groups(box_data.inputs)

        # Outputs
        box_model.add_outputs_groups(box_data.outputs)

        # Add default params value
        if hasattr(box_data, 'params') :
            for parameter, attributes in box_data.params.items() :
                # Copy parameter
                box_model._parameters[parameter] = copy.deepcopy(attributes)

        return box_model

    @staticmethod
    def fromDict(box_dict: dict):
        """Est la méthode appelée pour produire un modèle de box suivant un dictionnaire d'import."""

        # Get genral box informations
        library = box_dict['library']
        name = box_dict['name']

        #
        text = box_dict['text']

        # Get geometry
        position = QPointF(float(box_dict['geometry']['x']), float(box_dict['geometry']['y']))
        size = QSizeF(float(box_dict['geometry']['width']), float(box_dict['geometry']['height']))
        rotation = float(box_dict['geometry']['rotation'])

        # Build Box from BoxData
        box_data = BoxesLibrariesManager().get_box_data(library, name)
        
        # Handle Geometry value
        ## Size
        if size.width() < AbstractBoxModel.MinimunBoxWidth:
            size.setWidth(AbstractBoxModel.MinimunBoxWidth)
        if size.height() < AbstractBoxModel.MinimunBoxHeight:
            size.setHeight(AbstractBoxModel.MinimunBoxHeight)
        
        ## Rotation
        rotation = ((rotation // 90) % 4) * 90

        # Build Box Model
        box_model = BoxModel(
            library=box_data.box_library,
            name=box_data.box_name,
            position=position,
            size=size,
            rotation=rotation, can_be_loaded=True, text=text, icon=box_data.icon
        )
        
        box_model.box_type = box_data.box_type

        # Inputs
        box_model.add_inputs_groups(box_data.inputs, infinite_groups_reset=True)

        # Outputs
        box_model.add_outputs_groups(box_data.outputs, infinite_groups_reset=True)
        
        # Set Parameter values
        if hasattr(box_data, 'params') :
            for parameter, attributes in box_data.params.items() :
                # Copy parameter
                box_model._parameters[parameter] = copy.deepcopy(attributes)

                # Set with new value
                if parameter in box_dict["params"]:
                    box_model.set_parameter(parameter, box_dict["params"][parameter])


        return box_model


#-----------------------------------
# Imports
#-----------------------------------

from typing import overload
from PyQt5 import QtGui
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QSlider, QWidget

#-----------------------------------
# Class
#-----------------------------------

class RotationSlider (QSlider) :
    """Est l'élément graphique qui permet de modifier l'angle d'un élément graphique."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, parent : QWidget = None) :

        QSlider.__init__(self, parent)

        self.setMaximum(270)
        self.setMinimum(0)
        self.setTickInterval(90)
        self.setSingleStep(90)

    # -------------
    # Methods
    # -------------

    def change_value (self, pos_x : float) :
        old_value = (pos_x / self.geometry().width()) * (self.maximum() - self.minimum())
        value = 0

        if old_value < 45 :
            value = 0

        elif old_value < 135 :
            value = 90

        elif old_value < 225 :
            value = 180 

        else :
            value = 270

        self.setValue(int(value))

    def mousePressEvent (self, event : QtGui.QMouseEvent) -> None :
        self.change_value(event.pos().x())

    def mouseMoveEvent (self, event : QtGui.QMouseEvent) -> None :
        self.change_value(event.pos().x())

    def setValue (self, rotation) -> None :

        if isinstance(rotation, float) :
            super().setValue(int(rotation))

        elif isinstance(rotation, int) :
            super().setValue(rotation)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel

#-----------------------------------
# Functions
#-----------------------------------
def _update_box_value(box_model: BoxModel):
    box_model.setValue(box_model.get_parameter("gain")["value"])

def _function_amplifier_box (box, event_) -> List :
    
    v = box['gain'] * box.get_input(0).value
    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events



import matplotlib.pylab as plt
import numpy as np

C = [
        [1/2,   1/4,    1/8,    1/16,   1/32,   1/64],
        [4,     4/3,    4/6,    4/9,    4/12,   4/15],
        [1,     1/5,    1/10,   1/15,   1/20,   1/25]  
    ]


X = []
Values = []
Values.append([])
Values.append([])
Values.append([])

v = 0
while v < 0.05:
    X.append(v)
    v += 0.001

for i, x_ in enumerate(X):
    v = 0
    
    if x_ < 0.035 :
        Values[1].insert(0, 1)
    else :
        Values[1].insert(0, 0)

    if x_ < 0.003:
        Values[2].insert(0, 0)
    elif x_ < 0.014:
        Values[2].insert(0, 1)
    elif x_ < 0.028:
        Values[2].insert(0, 0)
    else:
        Values[2].insert(0, 1)

    if len(C) > 0:
        if len(C[0]) > 0 :
            v = C[0][0]
    
    for j in range(min(len(C[0])-1, len(Values[0]))):
        v += C[0][j + 1] * Values[0][j]

    input_count = len(Values) - 1
    coeff_count = len(C) - 1
    input_index_count = min(input_count, coeff_count)

    if x_ >= 0.013 and x_ <= 0.014 :
        print("oui")

    for input_index in range(min(coeff_count, input_count)) :

        input_value_index_count = len(C[input_index + 1])
        last_values_index_count = len(Values[input_index + 1])
        values_index_count = min(input_value_index_count, last_values_index_count)

        for value_index in range(values_index_count) :
            coeff = C[input_index + 1][value_index]
            val = Values[input_index + 1][value_index]
            v += coeff * val 
    
    Values[0].insert(0, v)

Y = []
I1 = []
I2 = []
for i, x_ in enumerate(X):
    Y.append(Values[0][len(X) - 1 - i])
    I1.append(Values[1][len(X) - 1 - i])
    I2.append(Values[2][len(X) - 1 - i])

fig = plt.figure('i1 & i2') 
plt.grid(True)
plt.title('i1 & i2')
plt.xlabel('Temps (s)')
plt.ylabel('Amplitude')
plt.plot(X, Values[1], X, Values[2])


fig2 = plt.figure('y')
plt.grid(True)
plt.title('y')
plt.xlabel('Temps (s)')
plt.ylabel('Amplitude')
plt.plot(X, Y)


plt.show()



#-----------------------------------
# Class
#-----------------------------------

class AbstractCodeModel :
    """Est la classe abstraite commune aux model des éléments graphique ayant à modifier du code."""

    # -------------
    # Constructors
    # -------------

    def __init__(self, code : str = '') :
        self._code : str = code
        self._base_code : str = code

    # -------------
    # Properties
    # -------------
    
    @property
    def editor_type (self) -> str :
        """Permet de récuperer le type d'éditeur à utiliser."""
        return 'code-editor'

    @property
    def code (self) -> str :
        """Permet de récuperer le code du planificateur en cours de modification."""
        return self._code

    @property
    def is_code_changed (self) -> bool :
        """Permet de savoir si le code en cours est différent du code "non sauvegarder"."""
        return not(self._code == self._base_code)

    # -------------
    # Methods
    # -------------

    def accept_changed (self) -> None :
        """Permet d'accepter les changements du code."""
        self._base_code : str = self._code


import time

import matplotlib.pylab as plt

from ..core.plci_core_scheduler import Scheduler
from ..core.plci_core_data_type import DataType
from ..core.plci_core_signal import Signal
from ..core.plci_core_scheduler_params import SchedulerParams

from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

print("Creation of the diagram")
# Scheduler parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(5.0, 0.001)

# Signals
scheduler.diagram.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 3, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

# Sources
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Sine', amplitude = 10, pulsation = 10, phase = 0))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Sine', amplitude = 5, pulsation = 20, phase = 0))

# Addition
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Math', 'Addition'))             

# Sink
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = 'Sinus', label_x = 'Temps (s)', label_y = 'Amplitude (v)' ))

# Sources additional
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Sine', amplitude = 1, pulsation = 100, phase = 0))
 
# Diagram creation

scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[4], scheduler.diagram.signals[3])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[0], scheduler.diagram.signals[1], scheduler.diagram.signals[3])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[2])

# Launch solver
simulate(scheduler)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Dict

from PyQt5.QtWidgets import QListWidget, QListWidgetItem

from muphyn.packages.interface.graphical_actions.abstract_graphical_action import AbstractGraphicalAction

#-----------------------------------
# Class
#-----------------------------------

class LibraryDialogRemoveAction (AbstractGraphicalAction) :
    """Est l'action défaisable pour supprimer des éléments dans la liste."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, lst : QListWidget, elements : List[str]) :
        self._lst = lst

        self._elements : Dict[str, int] = {}
        
        for element in elements :
            self._elements[element] = 0

    # -------------
    # Methods
    # -------------

    def do (self) :
        
        for element in self._elements :

            for i in range(self._lst.__len__()) :
                item = self._lst.item(i)

                if item.text() == element :
                    self._elements[element] = i
                    self._lst.takeItem(i)
                    break

    def undo (self) :
        
        for element in self._elements.__reversed__() :
            index = self._elements[element]
            QListWidgetItem(element)
            self._lst.insertItem(index, element)





# PyQt5 Imports
from PyQt5.QtCore import QPointF, QSizeF
from PyQt5.QtWidgets import QGraphicsItem

# Project Imports
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement

class AbstractComponentModel(AbstractMoveableGraphicalElement):

    def __init__(self, name: str, position: QPointF, size: QSizeF, rotation: float = 0.0, 
        text: str = '', parent: QGraphicsItem = None, symbol_path: str = None):

        AbstractMoveableGraphicalElement.__init__(name, position, size, rotation, text, parent)

        self._should_recompute = False



#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from typing import Iterable, Any

# PyQt5 Imports
from PyQt5.QtCore import QCoreApplication

# Project Imports
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.properties_pages.box_properties import BoxProperties
from muphyn.packages.interface.properties_pages.moveable_graphical_element_properties_editor import MoveableGraphicalElementPropertiesEditor
from muphyn.packages.interface.properties_pages.parameter_properties_editor import ParameterPropertiesEditor
from muphyn.packages.interface.properties_pages.signal_properties_editor import SignalPropertiesEditor
from muphyn.packages.interface.properties_pages.title_properties_element import TitlePropertiesElement
from muphyn.packages.interface.properties_pages.unknown_properties_editor import UnknownPropertiesEditor
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.properties_pages.infinite_input_properties_editor import InfiniteInputPropertiesEditor
from muphyn.packages.interface.properties_pages.infinite_output_properties_editor import InfiniteOutputPropertiesEditor

#-----------------------------------
# Function
#-----------------------------------

def get_properties_page (element : Any) -> Iterable[AbstractPropertiesEditor] :
    
    based_editor_loaded = False

    if isinstance(element, SignalLinkModel) : 
        yield SignalPropertiesEditor(element)
        return
    
    if isinstance(element, BoxOutputModel) :
        properties_page = None

    elif isinstance(element, BoxInputModel) : 
        properties_page = None

    elif isinstance(element, BoxModel) :
        box_model : BoxModel = element
        yield BoxProperties(box_model)

        # Inputs
        already_added_name_input = []
        for input_group in box_model.inputs_groups.values(): 
            if input_group.is_infinite and not(input_group.name in already_added_name_input):
                already_added_name_input.append(input_group.name)
                yield InfiniteInputPropertiesEditor(box_model, input_group)

        # Outputs
        already_added_name_output = []
        for output_group in box_model.outputs_groups.values(): 
            if output_group.is_infinite and not(output_group.name in already_added_name_output):
                    already_added_name_output.append(output_group.name)
                    yield InfiniteOutputPropertiesEditor(box_model, output_group)

        if box_model.get_parameters_len() > 0 :
            yield TitlePropertiesElement(QCoreApplication.translate('properties_page_builder', u"Parameters : ", None))
            for parameter in box_model.get_parameters() :
                yield ParameterPropertiesEditor(box_model, parameter)

        based_editor_loaded = True


    if based_editor_loaded == False :
        yield UnknownPropertiesEditor()
        return

    else :

        if isinstance(element, AbstractMoveableGraphicalElement) :
            yield MoveableGraphicalElementPropertiesEditor(element)
            


#-----------------------------------
# Imports
#-----------------------------------

from muphyn.packages.core.plci_core_data_type import DataType

#-----------------------------------
# Classs
#-----------------------------------

class AbstractSignalModel : 
    """Est la classe abstraite commune aux signaux typés dans l'interface."""

    # -------------
    # Constructors
    # -------------
    
    def __init__ (self, data_type : DataType) :

        self._data_type : DataType = data_type

    # -------------
    # Properties
    # -------------

    @property
    def data_type (self) -> DataType :
        """Permet de récuperer le type de l'élément."""
        return self._data_type

    @data_type.setter
    def data_type (self, data_type_ : DataType) -> None :
        """Permet de modifier le type de l'élément."""
        self._data_type = data_type_
        
    @property
    def is_connected_to_input (self) -> bool :
        """Permet de savoir si l'élément actuel est connecté à une entrée (ou est un entrée)."""
        ...


#-----------------------------------
# Imports
#-----------------------------------

from .plci_core_data_type import DataType
from typing import Any

#-----------------------------------
# Class
#-----------------------------------

class Data : 
    """Est l'objet contenant les données transité par les signaux."""

    # -------------
    # Constructors
    # -------------
    def __init__ (self, data_type_ : DataType, value_ : Any = None) :

        self._data_type = data_type_   

        if value_ is None :
            self._value = self._data_type.default_value()
        
        else :
            self._value = value_

    # -------------
    # Properties
    # -------------

    @property
    def data_type (self) -> DataType : 
        """Permet de récuperer le type de la donnée."""
        return self._data_type

    @property
    def value (self) -> Any :
        """Permet de récuperer la valeur actuelle."""
        return self._value

    def __str__ (self) -> str :
        """Permet de récuperer la données sous la forme d'un string."""
        return "Data [" + str(self.data_type) + " | value : " + str(self.value) + "]" 




#-----------------------------------
# Imports
#-----------------------------------

import importlib
import yaml
from typing import Dict

from .box_library_data import AbstractBoxData, CompositeBoxData, CodeBoxData
from .abstract_box_library_importer import AbstractBoxLibraryImporter

#-----------------------------------
# Class
#-----------------------------------

class BoxLibraryImporter_1_0_0 (AbstractBoxLibraryImporter) :
    """Est la classe V 1.0.0 d'importeur de boxes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        super().__init__()
        
    # -------------
    # Methods
    # -------------

    def import_box (self, path : str, file_name : str, absolute_yaml_file : str, boxes : Dict[str, AbstractBoxData]) -> Dict[str, AbstractBoxData] :
        """Est la méthode pour charger des boxes depuis des fichiers."""
        
        # Open File
        with open(absolute_yaml_file) as yaml_file_data :

            # Read YAML File
            file_data = yaml.load(yaml_file_data, Loader = yaml.FullLoader)

            if 'box' in file_data :
                # Get Box data                
                box_data = file_data['box']

                # Filter box based on the library name
                if box_data['library'].startswith('Boxes') :
                    # Build fill box name
                    library_name = box_data['library'] + "." + box_data['name']

                    # Si la box est déjà dans la bibliothèques 
                    if library_name in boxes :
                        # Et que sa version est plus haute.
                        if boxes[library_name].version >= box_data['version']:
                            return None
                    
                    if box_data['type'] == 'code' :
                        return {'library_name' : library_name, 'box_data' : self._import_code_box(library_name, box_data, path + '/' + file_name + '.py')}

                    elif box_data['type'] == 'composite' :
                        return {'library_name' : library_name, 'box_data' : self._import_composite_box(library_name, box_data)}

                elif box_data['library'].startswith('Electrical'):
                    
                    library_name = box_data['library'] + "." + box_data['name']


    def _import_composite_box (self, library_name, box_data) -> CompositeBoxData :
        """Permet de générer un instanciateur de box composite."""

        if box_data['inputs'] == 'None':
            inputs = []
        else :
            inputs = box_data['inputs']

        if box_data['outputs'] == 'None':
            outputs = []
        else :
            outputs = box_data['outputs']

        return CompositeBoxData (
            box_data['name'],
            box_data['library'],
            box_data['creator'],
            box_data['date_creation'],
            box_data['version'],
            box_data['signals'],
            box_data['boxes'],
            inputs,
            outputs,
        )   

    def _import_code_box (self, library_name, box_data, absolute_py_file) -> CodeBoxData :
        """Permet de générer un instanciateur de box de code."""

        # Importations des méthodes du fichier python.
        spec = importlib.util.spec_from_file_location(library_name, absolute_py_file)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)

        # Si la méthode init n'éxiste, pas création de celle-ci.
        if box_data['box_init_method'] == "None" : 
            init_method = lambda box, scheduler_params : ...
        else :
            init_method = getattr(foo, box_data['box_init_method'])

        # Si la fonction n'éxiste, pas création de celle-ci
        if box_data['box_function'] == "None" : 
            box_function = lambda box, event : ...
        else :
            box_function = getattr(foo, box_data['box_function'])

        # Si la méthode end n'éxiste, pas création de celle-ci
        if box_data['box_end_method'] == "None" : 
            end_method = lambda box : ...
        else :
            end_method = getattr(foo, box_data['box_end_method'])

        if box_data['inputs'] == 'None':
            inputs = []
        else :
            inputs = box_data['inputs']

        if box_data['outputs'] == 'None':
            outputs = []
        else :
            outputs = box_data['outputs']

        if box_data['params'] == 'None' :
            params = {}
        else :
            params = box_data['params']

        if "icon" not in box_data :
            icon = None
        else:
            icon = box_data['icon']

        # Retour de la nouvelle box de code.
        return CodeBoxData(
            box_name = box_data['name'],
            box_library = box_data['library'],
            wait_for_event = box_data['wait_for_events'],
            wait_for_all_signal_events = box_data['wait_for_all_signal_events'],
            params = params,
            box_init_method_ = init_method, 
            box_function_ = box_function,
            box_end_method_ = end_method, 
            creator = box_data['creator'],
            date_created = box_data['date_creation'],
            version = box_data['version'],
            inputs = inputs,
            outputs = outputs,
            icon = icon
        )